@@ -38,7 +38,9 @@
 #include <QRegExp>
 
 CppTree2Uml::CppTree2Uml(const QString& fileName, CodeImpThread* thread)
-  : m_thread(thread)
+  : m_thread(thread),
+    m_rootFolder(0),
+    m_doc(UMLApp::app()->document())
 {
     clear();
     QDir dir(fileName);
@@ -66,10 +68,35 @@ void CppTree2Uml::clear()
     m_anon = 0;
 }
 
+void CppTree2Uml::setRootPath(const QString &rootPath)
+{
+    m_rootPath = QDir::fromNativeSeparators(rootPath);
+    if (Settings::optionState().codeImportState.createArtifacts) {
+        if (!m_rootFolder) {
+            UMLFolder *componentView = m_doc->rootFolder(Uml::ModelType::Component);
+            if (!m_rootPath.isEmpty()) {
+                UMLFolder *root = Import_Utils::createSubDir(m_rootPath, componentView);
+                m_rootFolder = root;
+            } else {
+                m_rootFolder = componentView;
+            }
+        }
+    }
+}
+
 void CppTree2Uml::parseTranslationUnit(const ParsedFile &file)
 {
     clear();
-    Import_Utils::createArtifact(file.fileName(), 0, file->comment());
+    if (Settings::optionState().codeImportState.createArtifacts) {
+        QFileInfo fi(file.fileName());
+
+        UMLFolder *parent = m_rootFolder;
+        QString path = fi.path().replace(m_rootPath + QLatin1String("/"), QLatin1String(""));
+        if (!path.isEmpty())
+            parent = Import_Utils::createSubDir(path, m_rootFolder);
+
+        Import_Utils::createArtifact(fi.fileName(), parent, file->comment());
+    }
 
     TreeParser::parseTranslationUnit(file);
 }
@@ -94,9 +121,9 @@ void CppTree2Uml::parseNamespace(NamespaceAST* ast)
     if (m_thread) {
         m_thread->emitMessageToLog(QString(), QLatin1String("namespace ") + nsName);
     }
-    UMLObject *o = UMLApp::app()->document()->findUMLObject(nsName, UMLObject::ot_Package, m_currentNamespace[m_nsCnt]);
+    UMLObject *o = m_doc->findUMLObject(nsName, UMLObject::ot_Package, m_currentNamespace[m_nsCnt]);
     if (!o)
-        o = UMLApp::app()->document()->findUMLObject(nsName, UMLObject::ot_Class, m_currentNamespace[m_nsCnt]);
+        o = m_doc->findUMLObject(nsName, UMLObject::ot_Class, m_currentNamespace[m_nsCnt]);
     if (o && o->stereotype() == QLatin1String("class-or-package")) {
         o->setStereotype(QString());
         o->setBaseType(UMLObject::ot_Package);
@@ -170,8 +197,13 @@ void CppTree2Uml::parseTypedef(TypedefAST* ast)
                  Import_Utils::createUMLObject(UMLObject::ot_Datatype, id,
                                                m_currentNamespace[m_nsCnt]);
                 UMLDatatype *dt = typedefObj->asUMLDatatype();
-                dt->setIsReference();
-                dt->setOriginType(inner->asUMLClassifier());
+                if (dt) {
+                    dt->setIsReference();
+                    dt->setOriginType(inner->asUMLClassifier());
+                }
+                else {
+                    uError() << "Could not create datatype from" << id;
+                }
             } else {
                 Import_Utils::createUMLObject(UMLObject::ot_Class, id,
                                                m_currentNamespace[m_nsCnt],
@@ -268,20 +300,25 @@ void CppTree2Uml::parseFunctionDefinition(FunctionDefinitionAST* ast)
         return;
 
     bool isFriend = false;
-//:unused:    bool isVirtual = false;
+    bool isVirtual = false;
     bool isStatic = false;
-//:unused:    bool isInline = false;
+    bool isInline = false;
     bool isConstructor = false;
-
+    bool isDestructor = false;
+    bool isExplicit = false;
     bool isConstExpression = false;
 
     if (funSpec){
-//:unused:        QList<AST*> l = funSpec->nodeList();
-//:unused:        for (int i = 0; i < l.size(); ++i) {
-//:unused:            QString text = l.at(i)->text();
-//:unused:            if (text == "virtual") isVirtual = true;
-//:unused:            else if (text == "inline") isInline = true;
-//:unused:        }
+        QList<AST*> l = funSpec->nodeList();
+        for (int i = 0; i < l.size(); ++i) {
+            QString text = l.at(i)->text();
+            if (text == QLatin1String("virtual"))
+                isVirtual = true;
+            else if (text == QLatin1String("inline"))
+                isInline = true;
+            else if (text == QLatin1String("explicit"))
+                isExplicit = true;
+        }
     }
 
     if (storageSpec){
@@ -311,16 +348,38 @@ void CppTree2Uml::parseFunctionDefinition(FunctionDefinitionAST* ast)
     UMLOperation *m = Import_Utils::makeOperation(c, id);
     if (isConstExpression)
         m->setStereotype(QLatin1String("constexpr"));
-
+    if (isVirtual)
+        m->setVirtual(true);
+    if (isInline)
+        m->setInline(true);
+    if (d->override())
+        m->setOverride(true);
+    if (d->constant())
+        m->setConst(true);
     // if a class has no return type, it could be a constructor or
     // a destructor
-    if (d && returnType.isEmpty() && id.indexOf(QLatin1Char('~')) == -1)
-        isConstructor = true;
+    if (d && returnType.isEmpty()) {
+        if (id.indexOf(QLatin1Char('~')) == -1)
+            isConstructor = true;
+        else
+            isDestructor = true;
+    }
 
     parseFunctionArguments(d, m);
     Import_Utils::insertMethod(c, m, m_currentAccess, returnType,
-                               isStatic, false /*isAbstract*/, isFriend, isConstructor, m_comment);
+                               isStatic, false /*isAbstract*/, isFriend, isConstructor,
+                               isDestructor, m_comment);
     m_comment = QString();
+    if  (isConstructor) {
+        QString stereotype;
+        if (isExplicit)
+            stereotype.append(QLatin1String("explicit "));
+        if (isConstExpression)
+            stereotype.append(QLatin1String("constexpr "));
+        stereotype.append(QLatin1String("constructor"));
+        m->setStereotype(stereotype);
+    } else if (isConstExpression)
+        m->setStereotype(QLatin1String("constexpr"));
 
 /* For reference, Kdevelop does some more:
     method->setFileName(m_fileName);
@@ -356,18 +415,30 @@ void CppTree2Uml::parseClassSpecifier(ClassSpecifierAST* ast)
     if (m_thread) {
         m_thread->emitMessageToLog(QString(), QLatin1String("class ") + className);
     }
-    if (!scopeOfName(ast->name(), QStringList()).isEmpty()){
-        uDebug() << "skip private class declarations";
-        return;
+    QStringList scope = scopeOfName(ast->name(), QStringList());
+    UMLObject *localParent = 0;
+    if (!scope.isEmpty()) {
+        localParent = m_doc->findUMLObject(scope.join(QLatin1String("::")),
+                                           UMLObject::ot_Class, m_currentNamespace[m_nsCnt]);
+        if (!localParent)
+            localParent = m_doc->findUMLObject(scope.join(QLatin1String("::")),
+                                               UMLObject::ot_Package, m_currentNamespace[m_nsCnt]);
+        if (!localParent) {
+            localParent = Import_Utils::createUMLObject(UMLObject::ot_Class, className,
+                                                        m_currentNamespace[m_nsCnt],
+                                                        ast->comment(), QString(), true);
+            localParent->setStereotype(QLatin1String("class-or-package"));
+        }
+        m_currentNamespace[++m_nsCnt] = localParent->asUMLPackage();
     }
 
     if (className.isEmpty()) {
         className = QLatin1String("anon_") + QString::number(m_anon);
         m_anon++;
     }
-    UMLObject *o = UMLApp::app()->document()->findUMLObject(className, UMLObject::ot_Class, m_currentNamespace[m_nsCnt]);
+    UMLObject *o = m_doc->findUMLObject(className, UMLObject::ot_Class, m_currentNamespace[m_nsCnt]);
     if (!o)
-        o = UMLApp::app()->document()->findUMLObject(className, UMLObject::ot_Datatype, m_currentNamespace[m_nsCnt]);
+        o = m_doc->findUMLObject(className, UMLObject::ot_Datatype, m_currentNamespace[m_nsCnt]);
     if (o && o->stereotype() == QLatin1String("class-or-package")) {
         o->setStereotype(QString());
         o->setBaseType(UMLObject::ot_Class);
@@ -420,6 +491,8 @@ void CppTree2Uml::parseClassSpecifier(ClassSpecifierAST* ast)
     m_currentAccess = oldAccess;
     m_inSlots = oldInSlots;
     m_inSignals = oldInSignals;
+    if (localParent)
+        m_currentNamespace[m_nsCnt--] = 0;
 }
 
 void CppTree2Uml::parseEnumSpecifier(EnumSpecifierAST* ast)
@@ -507,7 +580,7 @@ void CppTree2Uml::parseDeclaration2(GroupAST* funSpec, GroupAST* storageSpec,
     }
 
     QString typeName = typeOfDeclaration(typeSpec, d);
-//:unused:    bool isFriend = false;
+    bool isFriend = false;
     bool isStatic = false;
 //:unused:    bool isInitialized = decl->initializer() != 0;
 
@@ -515,13 +588,18 @@ void CppTree2Uml::parseDeclaration2(GroupAST* funSpec, GroupAST* storageSpec,
         QList<AST*> l = storageSpec->nodeList();
         for (int i = 0; i < l.size(); ++i) {
             QString text = l.at(i)->text();
-            if (text == QLatin1String("static")) isStatic = true;
-//:unused:            else if (text == QLatin1String("friend")) isFriend = true;
+            if (text == QLatin1String("static"))
+                isStatic = true;
+            else if (text == QLatin1String("mutable"))
+                typeName.prepend(text + QLatin1String(" "));
+            else if (text == QLatin1String("friend"))
+                isFriend = true;
         }
     }
 
-    Import_Utils::insertAttribute(c, m_currentAccess, id, typeName,
-                                  m_comment, isStatic);
+    UMLAttribute *attribute = Import_Utils::insertAttribute(c, m_currentAccess, id, typeName, m_comment, isStatic);
+    if (isFriend)
+        attribute->setStereotype(QLatin1String("friend"));
     m_comment = QString();
 }
 
@@ -541,20 +619,26 @@ void CppTree2Uml::parseFunctionDeclaration(GroupAST* funSpec, GroupAST* storageS
                                              TypeSpecifierAST * typeSpec, InitDeclaratorAST * decl)
 {
     bool isFriend = false;
-//:unused:    bool isVirtual = false;
+    bool isVirtual = false;
     bool isStatic = false;
-//:unused:    bool isInline = false;
+    bool isInline = false;
     bool isPure = decl->initializer() != 0;
     bool isConstructor = false;
     bool isConstExpression = false;
+    bool isDestructor = false;
+    bool isExplicit = false;
 
     if (funSpec){
-//:unused:        QList<AST*> l = funSpec->nodeList();
-//:unused:        for (int i = 0; i < l.size(); ++i) {
-//:unused:            QString text = l.at(i)->text();
-//:unused:            if (text == QLatin1String("virtual")) isVirtual = true;
-//:unused:            else if (text == QLatin1String("inline")) isInline = true;
-//:unused:        }
+        QList<AST*> l = funSpec->nodeList();
+        for (int i = 0; i < l.size(); ++i) {
+            QString text = l.at(i)->text();
+            if (text == QLatin1String("virtual"))
+                isVirtual = true;
+            else if (text == QLatin1String("inline"))
+                isInline = true;
+            else if (text == QLatin1String("explicit"))
+                isExplicit = true;
+        }
     }
 
     if (storageSpec){
@@ -579,18 +663,42 @@ void CppTree2Uml::parseFunctionDeclaration(GroupAST* funSpec, GroupAST* storageS
 
     QString returnType = typeOfDeclaration(typeSpec, d);
     UMLOperation *m = Import_Utils::makeOperation(c, id);
+    if (d->override())
+        m->setOverride(true);
+    if (d->constant())
+        m->setConst(true);
     if (isConstExpression)
         m->setStereotype(QLatin1String("constexpr"));
+    if (isVirtual)
+        m->setVirtual(true);
+    if (isInline)
+        m->setInline(true);
     // if a class has no return type, it could de a constructor or
     // a destructor
-    if (d && returnType.isEmpty() && id.indexOf(QLatin1Char('~')) == -1)
-        isConstructor = true;
+    if (d && returnType.isEmpty()) {
+        if (id.indexOf(QLatin1Char('~')) == -1)
+            isConstructor = true;
+        else
+            isDestructor = true;
+    }
 
     parseFunctionArguments(d, m);
     Import_Utils::insertMethod(c, m, m_currentAccess, returnType,
-                               isStatic, isPure, isFriend, isConstructor, m_comment);
+                               isStatic, isPure, isFriend, isConstructor, isDestructor, m_comment);
     if  (isPure)
         c->setAbstract(true);
+
+    if  (isConstructor) {
+        QString stereotype;
+        if (isExplicit)
+            stereotype.append(QLatin1String("explicit "));
+        if (isConstExpression)
+            stereotype.append(QLatin1String("constexpr "));
+        stereotype.append(QLatin1String("constructor"));
+        m->setStereotype(stereotype);
+    } else if (isConstExpression)
+        m->setStereotype(QLatin1String("constexpr"));
+
     m_comment = QString();
 }
 
